#ifndef _SESSION_H
#define _SESSION_H

#ifdef HAVE_SSL
#include "libssl.h"
#endif
#include "serverconfig.h"
#include "liblist.h"

#define TIMER_OFF -1
#define MAX_CHILD_ID 30000
#define OUTPUT_BUFFER_SIZE 2000

#define fb_FILESYSTEM 0
#define fb_ACCESSLIST 1
#define fb_NO_CGI     2
#define fb_SYMLINK    3
#define fb_DENYBOT    4

typedef struct type_headerfield {
	char *data;
	int  value_offset;

	struct type_headerfield *next;
} t_headerfield;

#ifdef HAVE_SSL
enum t_cause_of_301 { missing_slash, require_ssl };
#endif

typedef struct type_session {
	t_config       *config;

	int            client_id;
	int            client_socket;
	t_binding      *binding;
	bool           socket_open;
	bool           head_request;
	bool           keep_alive;
	int            kept_alive;
	bool           is_cgi_script;
	char           *cgi_handler;
	char           *request, *method, *uri, *vars, *http_version, *body, *file_on_disk;
	long           header_length, content_length, buffer_size, bytes_in_buffer;
	char           *local_user;
	bool           body_copied;
	bool           vars_copied;
	bool           force_quit;
	bool           uri_is_dir;
	bool           accept_gzip;
	bool           encode_gzip;
	t_headerfield  *headerfields;
	long           ip_address;
	char           *mimetype;
	t_host         *host;
	bool           host_copied;
	char           *remote_user;
	t_directory    *directory;
	bool           handling_error;
	char           *reason_for_403;
	char           *cookie;
	bool           cookie_copied;
	bool           websiteroot_copied;
	long           bytes_sent;
	int            return_code;
	int            error_code;

	/* Throttling: send_buffer() in send.c
	 */
	long           throttle;
	long           bytecounter;
	int            throttle_timer;
	bool           part_of_dirspeed;

	/* Flooding protection
	 */
	int            flooding_timer;
	int            flooding_counter;

	/* SSL
	 */
#ifdef HAVE_SSL
	SSL            *ssl_data;
	enum t_cause_of_301 cause_of_301;
#endif

	/* Output buffer
	 */
	char           output_buffer[OUTPUT_BUFFER_SIZE];
	int            output_size;
} t_session;

void init_session(t_session *session);
void reset_session(t_session *session);
//void destroy_session(t_session *session);

void set_403_reason(t_session *session, int reason);
t_headerfield *parse_headerfields(char *line);
char *get_headerfield(char *key, t_headerfield *headerfields);
void remove_headerfields(t_session *session);

int  get_homedir(t_session *session, char *username);
bool duplicate_host(t_session *session);
void restore_host(t_session *session);
bool is_volatile_object(t_session *session);
int  copy_directory_settings(t_session *session);
int  set_environment(t_session *session, char *buffer, int max);
char *file_in_chroot(t_session *session, char *file);
bool client_is_rejected_bot(t_session *session);
void close_socket(t_session *session);

#endif
